
/* Pure Load Balancer - (C)opyleft 2003 Jedi/Sector One <j@pureftpd.org> */

#include <config.h>
#include "plb.h"
#include "plb_globals.h"
#ifdef WITH_DMALLOC
# include <dmalloc.h>
#endif

#ifdef USE_EEL
eel_cb_return_value server_forward_request_timeout(eel_group_id gid, void *client_, int fd)
{
    Client * const client = client_;
    
    handle_timeout(gid, client);
    return EEL_ERROR;
}
eel_cb_return_value server_forward_request(eel_group_id gid, void *client_, int fd)
{
    Client * const client = client_;
    ssize_t written;

    if (client->postbuf == NULL) {
        client_disconnect(gid, client);
        return EEL_ERROR;
    }    
    if (client->postbuf_len <= (size_t) 0U) {
        goto already_written;
    }
    if ((written = write(client->server_fd,
                         client->postbuf + client->postbuf_written,
                         client->postbuf_len - client->postbuf_written)) !=
        (ssize_t) (client->postbuf_len - client->postbuf_written)) {
        if (written <= (ssize_t) 0) {
            if (errno == EINTR) {
                eel_add_write_timeout(gid,
                  server_forward_request,
                  server_forward_request_timeout,
                  client, 
                  fd, 
                  TIMEVAL_TO_MILLISECONDS(timeout_forward_server_write));
                return EEL_SUCCESS;
            }
            plb_log(LL_NOTIFY,
                    "Error while sending a forwarded request : [%s]",
                    strerror(errno));
            client_disconnect(gid, client);
            return EEL_ERROR;
        }
        client->postbuf_written += written;
        eel_add_write_timeout(gid,
                  server_forward_request,
                  server_forward_request_timeout,
                  client, 
                  fd, 
                  TIMEVAL_TO_MILLISECONDS(timeout_forward_server_write));
        return EEL_SUCCESS;
    }
    client->postbuf_len = (size_t) 0U;    
    already_written:
    return EEL_SUCCESS;
}
#else
void server_forward_request(int evfd, short event, void *client_)
{
    Client * const client = client_;
    ssize_t written;

    (void) evfd;
    if (handle_timeout(event, client) != 0) {
        return;
    }
    if (client->postbuf == NULL) {
        client_disconnect(client);
        return;
    }    
    if (client->postbuf_len <= (size_t) 0U) {
        goto already_written;
    }
    if ((written = write(client->server_fd,
                         client->postbuf + client->postbuf_written,
                         client->postbuf_len - client->postbuf_written)) !=
        (ssize_t) (client->postbuf_len - client->postbuf_written)) {
        if (written <= (ssize_t) 0) {
            if (errno == EINTR) {
                event_add(&client->server_write_ev,
                          &timeout_forward_server_write);
                return;
            }
            plb_log(LL_NOTIFY,
                    "Error while sending a forwarded request : [%s]",
                    strerror(errno));
            client_disconnect(client);
            return;
        }
        client->postbuf_written += written;
        event_add(&client->server_write_ev, &timeout_forward_server_write);
        return;
    }
    client->postbuf_len = (size_t) 0U;    
    already_written:
    (void) client;
}
#endif

#ifdef USE_EEL
void client_forward_request(eel_group_id gid, Client * const client)
{
    ssize_t readen;
    
    if (client->postbuf == NULL) {
        if ((client->postbuf = malloc(POSTBUF_CHUNKS)) == NULL) {
            plb_log(LL_ERROR,
                    "Out of memory for the forwarding buffer : [%s]",
                    strerror(errno));
            client_disconnect(gid, client);            
            return;
        }
        client->postbuf_len = (size_t) 0U;
        client->postbuf_written = (size_t) 0U;        
    }
    if ((readen = read(client->client_fd, client->postbuf, POSTBUF_CHUNKS))
        <= (ssize_t) 0) {
        if (readen < (ssize_t) 0) {
            if (errno == EINTR) {
                return;
            }
            plb_log(LL_NOTIFY,
                    "Error while reading to-be-forwarded data : [%s]",
                    strerror(errno));
        }        
        client_disconnect(gid, client);
        return;        
    }
    client->postbuf_len = (size_t) readen;
    client->postbuf_written = (size_t) 0U;
    
    // *** new/different gid for server?? or AND?? ---> see client_read_header/connect_server
    // they do the same thing, only separately. I think this add write for the server would 
    // replace connect_server's in libevent...not sure what it will do in eel
    eel_add_write_timeout(gid,
      server_forward_request,
      server_forward_request_timeout,
      client,
      client->server_fd,
      TIMEVAL_TO_MILLISECONDS(timeout_forward_server_write));
    eel_add_read_timeout(gid,
      client_read,
      client_read_timeout,
      client,
      client->client_fd,
      TIMEVAL_TO_MILLISECONDS(timeout_forward_client_read));
}
#else
void client_forward_request(Client * const client)
{
    ssize_t readen;
    
    if (client->postbuf == NULL) {
        if ((client->postbuf = malloc(POSTBUF_CHUNKS)) == NULL) {
            plb_log(LL_ERROR,
                    "Out of memory for the forwarding buffer : [%s]",
                    strerror(errno));
            client_disconnect(client);            
            return;
        }
        client->postbuf_len = (size_t) 0U;
        client->postbuf_written = (size_t) 0U;        
    }
    if ((readen = read(client->client_fd, client->postbuf, POSTBUF_CHUNKS))
        <= (ssize_t) 0) {
        if (readen < (ssize_t) 0) {
            if (errno == EINTR) {
                return;
            }
            plb_log(LL_NOTIFY,
                    "Error while reading to-be-forwarded data : [%s]",
                    strerror(errno));
        }        
        client_disconnect(client);
        return;        
    }
    client->postbuf_len = (size_t) readen;
    client->postbuf_written = (size_t) 0U;
    
    event_set(&client->server_write_ev, client->server_fd, EV_WRITE,
              server_forward_request, client);
    event_add(&client->server_write_ev, &timeout_forward_server_write);
    event_add(&client->client_read_ev, &timeout_forward_client_read);
}
#endif

#ifdef USE_EEL
eel_cb_return_value server_forward_reply_timeout(eel_group_id gid, void *client_, int fd)
{
    Client * const client = client_;    
    
    handle_timeout(gid, client);
    return EEL_ERROR;
}
eel_cb_return_value server_forward_reply(eel_group_id gid, void *client_, int fd)
{
    Client * const client = client_;    
    ssize_t readen;

    if (client->replybuf == NULL) {
        if ((client->replybuf = malloc(REPLYBUF_CHUNKS)) == NULL) {
            plb_log(LL_ERROR,
                    "Out of memory when trying to get a reply buffer : [%s]",
                    strerror(errno));
            client_disconnect(gid, client);
            return EEL_ERROR;
        }
        client->replybuf_len = (size_t) 0U;
        client->replybuf_written = (size_t) 0U;        
    }
    if ((readen = read(client->server_fd, client->replybuf, REPLYBUF_CHUNKS))
        <= (ssize_t) 0) {
        if (errno == EINTR) {
            eel_add_read_timeout(gid,
              server_forward_reply,
              server_forward_reply_timeout,
              client,
              client->server_fd,
              TIMEVAL_TO_MILLISECONDS(timeout_forward_server_read));
            return EEL_SUCCESS;
        }
        if (readen < (ssize_t) 0) {
            plb_log(LL_NOTIFY,
                    "Error while reading the data to be forwarded : [%s]",
                    strerror(errno));
        }
        client_disconnect(gid, client);
        return EEL_ERROR;        
    }
    client->replybuf_len = (size_t) readen;
    client->replybuf_written = (size_t) 0U;
    
    eel_add_write_timeout(gid,
              client_forward_reply,
              client_forward_reply_timeout,
              client,
              client->server_fd,
              TIMEVAL_TO_MILLISECONDS(timeout_forward_client_write));
    return EEL_SUCCESS;
}
#else
void server_forward_reply(int evfd, short event, void *client_)
{
    Client * const client = client_;    
    ssize_t readen;

    (void) evfd;
    if (handle_timeout(event, client) != 0) {
        return;
    }
    if (client->replybuf == NULL) {
        if ((client->replybuf = malloc(REPLYBUF_CHUNKS)) == NULL) {
            plb_log(LL_ERROR,
                    "Out of memory when trying to get a reply buffer : [%s]",
                    strerror(errno));
            client_disconnect(client);
            return;
        }
        client->replybuf_len = (size_t) 0U;
        client->replybuf_written = (size_t) 0U;        
    }
    if ((readen = read(client->server_fd, client->replybuf, REPLYBUF_CHUNKS))
        <= (ssize_t) 0) {
        if (errno == EINTR) {
            event_add(&client->server_read_ev, &timeout_forward_server_read);
            return;
        }
        if (readen < (ssize_t) 0) {
            plb_log(LL_NOTIFY,
                    "Error while reading the data to be forwarded : [%s]",
                    strerror(errno));
        }
        client_disconnect(client);
        return;        
    }
    client->replybuf_len = (size_t) readen;
    client->replybuf_written = (size_t) 0U;
    
    event_set(&client->client_write_ev, client->client_fd, EV_WRITE,
              client_forward_reply, client);
    event_add(&client->client_write_ev, &timeout_forward_client_write);
}
#endif

#ifdef USE_EEL
eel_cb_return_value client_forward_reply_timeout(eel_group_id gid, void *client_, int fd)
{
    Client * const client = client_;    
    
    handle_timeout(gid, client);
    return EEL_ERROR;
}
eel_cb_return_value client_forward_reply(eel_group_id gid, void *client_, int fd)
{
    Client * const client = client_;
    ssize_t written;

    if (client->replybuf == NULL) {
        client_disconnect(gid, client);
        return EEL_ERROR;
    }
    if (client->replybuf_len <= (size_t) 0U) {
        goto already_written;
    }    
    if ((written = write(client->client_fd, 
                         client->replybuf + client->replybuf_written,
                         client->replybuf_len - client->replybuf_written)) !=
        (ssize_t) (client->replybuf_len - client->replybuf_written)) {        
        if (written <= (ssize_t) 0) {
            if (errno == EINTR) {
                eel_add_write_timeout(gid,
                  client_forward_reply,
                  client_forward_reply_timeout,
                  client,
                  client->client_fd,
                  TIMEVAL_TO_MILLISECONDS(timeout_forward_client_write));
                return EEL_SUCCESS;
            }
            plb_log(LL_NOTIFY, "Error while sending content data : [%s]",
                    strerror(errno));
            client_disconnect(gid, client);
            return EEL_ERROR;
        }
        client->replybuf_written += written;
        eel_add_write_timeout(gid,
                  client_forward_reply,
                  client_forward_reply_timeout,
                  client,
                  client->client_fd,
                  TIMEVAL_TO_MILLISECONDS(timeout_forward_client_write));
        return EEL_SUCCESS;                
    }
    client->replybuf_len = (size_t) 0U;    
    already_written:    
    eel_add_read_timeout(gid,
                  server_forward_reply,
                  server_forward_reply_timeout,
                  client,
                  client->server_fd,
                  TIMEVAL_TO_MILLISECONDS(timeout_forward_server_read));
    return EEL_SUCCESS;
}
#else
void client_forward_reply(int evfd, short event, void *client_)
{
    Client * const client = client_;
    ssize_t written;

    (void) evfd;
    if (handle_timeout(event, client) != 0) {
        return;
    }
    if (client->replybuf == NULL) {
        client_disconnect(client);
        return;
    }
    if (client->replybuf_len <= (size_t) 0U) {
        goto already_written;
    }    
    if ((written = write(client->client_fd, 
                         client->replybuf + client->replybuf_written,
                         client->replybuf_len - client->replybuf_written)) !=
        (ssize_t) (client->replybuf_len - client->replybuf_written)) {        
        if (written <= (ssize_t) 0) {
            if (errno == EINTR) {
                event_add(&client->client_write_ev,
                          &timeout_forward_client_write);
                return;
            }
            plb_log(LL_NOTIFY, "Error while sending content data : [%s]",
                    strerror(errno));
            client_disconnect(client);
            return;
        }
        client->replybuf_written += written;
        event_add(&client->client_write_ev, &timeout_forward_client_write);
        return;                
    }
    client->replybuf_len = (size_t) 0U;    
    already_written:    
    event_set(&client->server_read_ev, client->server_fd, EV_READ,
              server_forward_reply, client);
    event_add(&client->server_read_ev, &timeout_forward_server_read);
}
#endif


